import can
from cantools import database
from configparser import ConfigParser
import logging
from .node import Node

log = logging.getLogger(__name__)



class Network:
    def __init__(self,bus_path:str=None, bus:can.BusABC = None, database_path:str = None) -> None:
        self._bus = can.interface.Bus()
        listener = NodeListener(self)
        config = ConfigParser().read('can.ini')
        self._loop = None
        self._notifier = can.Notifier(self._bus, (listener,), 1, self._loop)
        self._database = database.load_file(database_path or config.get('database','path'))
        for node_name in self._database.nodes:
            self._create_node(node_name.name)

    def _create_node(self, name:str) -> None:
        """Create a node"""
        self.__dict__.update({name: Node(name, self._bus, self._database)})
        return self.__dict__[name]
    
    @property
    def nodes(self) -> list[Node]:
        """Get all nodes"""
        nodes=[]
        for node in self._database.nodes:
            nodes.append(self.__dict__[node.name])
        return nodes



class NodeListener(can.Listener):
    def __init__(self, network: Network):
        self._network = network

    def on_message_received(self, msg: can.Message):
        if msg.is_error_frame or msg.is_remote_frame or msg.is_fd:
            # rtr is currently not supported
            return

        with self._network.lock:
            callback = self._network.subscriptions.get(msg.arbitration_id, None)

        if not callback:
            return

        try:
            callback(msg.arbitration_id, msg.data)
        except Exception as e:
            log.exception(f"{e!r} while processing {msg!r}")


class Listener(can.Listener):

    def __init__(self, database, messages, input_queue, on_message):
        self._database = database
        self._messages = messages
        self._input_queue = input_queue
        self._on_message = on_message

    def on_message_received(self, msg):
        if msg.is_error_frame or msg.is_remote_frame:
            return

        try:
            database_message = self._database.get_message_by_frame_id(
                msg.arbitration_id)
        except KeyError:
            return

        if database_message.name not in self._messages:
            return

        message = self._messages[database_message.name]

        if not message.enabled:
            return

        decoded =    database_message.decode(msg.data,
                                                         message.decode_choices,
                                                         message.scaling)

        if self._on_message:
            self._on_message(decoded)

        self._input_queue.put(decoded)



import can
from cantools.database import Database
import logging
from .message import create_message
         

class Node:
    """ControlUnit (CU)"""
    def __init__(self, name:str, bus:can.BusABC, database:Database) -> None:
        self._name = name
        self._bus = bus
        self._database = database
        listener = NodeListener(self)
        self._notifier = can.Notifier(self._bus, [listener])
        self.__create_signal_values(database, name)

    def __create_signal_values(self, database:Database, node_name:str=None) -> None:
        """Create Value instances for all signals in object"""
        logging.debug('Create signal values for %s', self)
        for message in database.messages:
            logging.debug(message.senders)
            if node_name in message.senders:
                logging.debug('Create sender message %s', message._name)
                setattr(self, message._name, create_message(message, True, self._bus ) )
            else:
                for signal in message._signals:
                    is_in_node_as_reveiver = False
                    if node_name in signal.receivers:
                        is_in_node_as_reveiver = True
                if is_in_node_as_reveiver:
                    logging.debug('Create receiver message %s', message._name)
                    setattr(self, message._name, create_message(message, False) )
        
    def _on_message(self, message:can.Message) -> None:
        """Callback for received CAN messages"""
        print(message)
    def _send_message(self, message:can.Message) -> None:
        """Callback for received CAN messages"""
        logging.debug('Send CAN message from Node %s: %s', self._name,message)
        self._bus.send(message)


class NodeListener(can.Listener):
    def __init__(self, node: Node) -> None:
        self._node = node

    def on_message_received(self, msg: can.Message):
        if msg.is_error_frame or msg.is_remote_frame: # or msg.is_fd: #todo what is 
            # rtr is currently not supported
            return
        try:
            database_message = self._node._database.get_message_by_frame_id(
                msg.arbitration_id)
        except KeyError:
            logging.ERROR('Received unknown message with arbitration id {}'.format( msg.arbitration_id ) )
            return
        try:
            self._node.__dict__[database_message._name]._update_data(msg)
        except:
            logging.ERROR('Received message with arbitration id {} and is not a instance of this node. '.format( msg.arbitration_id ) )



import unittest
import can
from canbro.node import Node

class TestNode(unittest.TestCase):
    def setUp(self):
        self.bus = can.interface.Bus(bustype='virtual')
        self.db = can.Database()
        self.db.add_message(can.Message(arbitration_id=0x100, data=[0, 1, 2, 3]))
        self.node = Node(name='test_node', bus=self.bus, database=self.db)

    def tearDown(self):
        self.bus.shutdown()

    def test_create_signal_values(self):
        self.assertEqual(len(self.node.__dict__), 1)
        self.assertTrue(hasattr(self.node, '0x100'))

    def test_on_message(self):
        msg = can.Message(arbitration_id=0x100, data=[4, 5, 6, 7])
        self.node._on_message(msg)
        # TODO: assert that the message was printed

    def test_send_message(self):
        msg = can.Message(arbitration_id=0x100, data=[4, 5, 6, 7])
        self.node._send_message(msg)
        # TODO: assert that the message was sent on the bus


import can
from cantools import database  
from cantools.autosar import compute_profile2_crc
from .e2e import E2E
from broqer import Value, op, Sink
import logging
import types
import asyncio
import datetime
import ast


class Signal(Value):
    """
    A class representing a signal in a message.

    Attributes:
    - metadata: An object containing metadata about the signal.
    """
    def __init__(self, metadata):
        super().__init__()
        self._metadata = metadata
    def init_value(self):
        if self._metadata.initial != None:
            self._state = self._metadata.initial
            logging.debug("set initial value of signal {} to {} with type {}".format(self._metadata.name,self._state,type(self._state)))
        else:
            logging.error("signal {} has no initial value".format(self._metadata.name,))



class Message(Value):
    """
    Represents a CAN message.

    Attributes:
        _metadata (database.Message): The metadata of the message.
        e2e (E2E): The end-to-end data object.
    """

    def __init__(self, metadata:database.Message, data_ids:list=None):
        """
        Initializes a new instance of the Message class.

        Args:
            metadata (database.Message): The metadata of the message.
            data_ids (list, optional): The list of data IDs. Defaults to None.
        """
        super().__init__()
        self._metadata = metadata
        if data_ids is not None:
            self.e2e = E2E()
            logging.info("set data ids, but just autosar profile 2 is supported!")
            self.e2e.data_ids = data_ids
        else:
            self.e2e = None
        logging.debug("create message {}".format(metadata._name))

    def add_data_ids(self, data_ids:list, crc_signal:str, snc_signal:str):
        """
        Adds data IDs to the message.

        Args:
            data_ids (list): The list of data IDs.
            crc_signal (str): The name of the CRC signal.
            snc_signal (str): The name of the SNC signal.
        """
        if self.e2e is None:
            self.e2e = E2E()
            logging.info("set data ids, but just autosar profile 2 is supported!")
        self.e2e.data_ids = data_ids
        self.e2e.crc_signal_name = crc_signal
        self.e2e.snc_signal_name = snc_signal
         

class MessageTx(Message):
    """
    A class representing a CAN message that can be transmitted on a CAN bus.

    Args:
        metadata (database.Message): The metadata associated with the message.
        can_bus (can.BusABC, optional): The CAN bus to transmit the message on. Defaults to None.

    Attributes:
        _can_bus (can.BusABC): The CAN bus to transmit the message on.
        _send_msg (bool): A flag indicating whether the message should be sent after it is updated.
    """

    def __init__(self, metadata:database.Message, can_bus:can.BusABC=None):
        super().__init__(metadata)
        self._can_bus = can_bus
        self._send_msg = False
        for signal in metadata._signals:
            setattr(self, "_signal_"+signal.name, Signal(signal) )
            setattr(self, "_set_"+signal.name, types.MethodType(lambda self,value: self.__dict__["_signal_"+signal.name].notify(value),self) )
            #if signal.initial == None:
            #    logging.error("signal {} has no initial value".format(signal.name))
            #if signal.initial != None: #todo init value
            #    self.__dict__["_signal_"+signal.name]._state = signal.initial
            setattr(self, "_get_"+signal.name, types.MethodType(lambda self: self.__dict__["_signal_"+signal.name].get() ,self )  )
            self.__dict__["_signal_"+signal.name].subscribe(Sink( self._update_can_message))   
    
    def _get_signals(self) -> dict: 
        """
        Returns a dictionary containing the current values of all signals in the message.

        Returns:
            dict: A dictionary containing the current values of all signals in the message.
        """
        data = dict()
        for signal in self._metadata._signals:
            data[signal.name] = self.__dict__["_signal_"+signal.name].get()
        return data
    
    def send_after_update(self, value:bool = True) -> None:
        """
        Sets the flag indicating whether the message should be sent after it is updated.

        Args:
            value (bool, optional): The value to set the flag to. Defaults to True.
        """
        self._send_msg = value
    
    def send(self) -> None:
        """
        Sends the message on the associated CAN bus.
        """
        self._can_bus.send(self._can_message)
    
    def _update_message(self) -> None:
        """
        Updates the message with the current values of all signals.
        """
        logging.debug('Update CAN message: %s', self._metadata._name)
        arbitration_id = self._metadata.frame_id
        extended_id = self._metadata.is_extended_frame
        data = self._get_signals()
        logging.debug('Update CAN Data: {}'.format(data) )
        pruned_data = self._metadata.gather_signals(data)
        data = self._metadata.encode(pruned_data)
        self._can_message = can.Message(arbitration_id=arbitration_id, is_extended_id=extended_id, data=data)
        self.notify(self._can_message)

    def _update_can_message(self,value) -> None:
        """
        Updates the message and sends it on the associated CAN bus if the send flag is set.

        Args:
            value: The new value of a signal in the message.
        """
        self._update_message()
        if self._send_msg:
            self._can_bus.send(self._can_message)


class MessageTxCycle(MessageTx):
    """
    A class representing a periodic message that is sent cyclically on the CAN bus.

    Attributes:
    metadata (database.Message): The metadata of the message.
    can_bus (can.BusABC): The CAN bus to send the message on.
    _periodic_task: The periodic task that sends the message.
    _send_msg (bool): A flag indicating whether the message should be sent.
    """

    def __init__(self, metadata:database.Message, can_bus:can.BusABC=None):
        super().__init__(metadata, can_bus)
        self._periodic_task = None
        self._send_msg = False
    
    def send_after_update(self, value):
        """
        Raises an error as a periodic message cannot be sent after an update.
        """
        logging.ERROR("is a periodic message and can not be send after update")
    
    def _update_can_message(self,value) -> None:
        """
        Updates the CAN message with the current signal values.
        """
        self._update_message()    
        if self._periodic_task is not None:
            self._periodic_task.modify_data(self._can_message)

    def start_periodic(self):
        """
        Starts sending the message periodically on the CAN bus.
        """
        if self._metadata.cycle_time:
            for signal in self._metadata._signals:
                self.__dict__["_signal_"+signal.name].init_value()
            #self._periodic_publisher = self | op.Throttle(self._metadata.cycle_time / 1000.0)
            self._update_can_message(None)
            self._periodic_task = self._can_bus.send_periodic( self._can_message, self._metadata.cycle_time / 1000.0)
            logging.debug("set periodic publisher for message {}".format(self._metadata._name))
            #self._periodic_publisher.subscribe(self._send_msg)
            #self._periodic_publisher.subscribe(self.notify)
            self.running = True
        else:
            print("is not a periodic message")

    def stop_periodic(self):
        """
        Stops sending the message periodically on the CAN bus.
        """
        if self._periodic_task is not None:
            self._periodic_task.stop()
            self._periodic_task = None
        #if self._metadata.cycle_time:
            #self._periodic_publisher.unsubscribe(self._send_msg)
            #self._periodic_publisher.unsubscribe(self.notify)
            self.running = False
        else:
            print("is not a periodic message")


class MessageTxCycleE2E(MessageTxCycle):
    """
    A class representing a periodic message with end-to-end (E2E) protection.

    Attributes:
    - running (bool): Indicates whether the periodic message is currently running.
    - metadata (database.Message): The metadata of the message.
    - can_bus (can.BusABC): The CAN bus to which the message is sent.
    - crc_signal (str): The name of the signal used for cyclic redundancy check (CRC) computation.
    - snc_signal (str): The name of the signal used for sequence number counter (SNC) computation.
    - data_ids (list): The list of data IDs used for E2E protection.
    - timestamp (float): The timestamp of the last message sent.

    Methods:
    - __init__(self, metadata:database.Message, can_bus:can.BusABC): Initializes the MessageTxCycleE2E object.
    - _update_e2e(self, msg: can.Message) -> None: Updates the E2E protection of the message.
    - _update_can_message(self, value) -> None: Updates the CAN message.
    - __cycle(self, value): Sends the periodic message and returns the SNC.
    - _increase_counter(self, snc:int): Increases the SNC and notifies subscribers.
    - start_periodic(self): Starts the periodic message.
    - stop_periodic(self): Stops the periodic message.
    """
    running = False
    def __init__(self, metadata:database.Message, can_bus:can.BusABC):
        """
        Initializes the MessageTxCycleE2E object.

        Args:
        - metadata (database.Message): The metadata of the message.
        - can_bus (can.BusABC): The CAN bus to which the message is sent.
        """
        super().__init__(metadata, can_bus)
        for signal in metadata.signals:
            sigType = 'GenSigFuncType'
            if sigType in signal.dbc.attributes:
                if signal.dbc.attributes[sigType].value == 2:
                    crc_signal = signal.name
                    for subscriber_obj in self.__dict__["_signal_"+crc_signal].subscriptions:
                        self.__dict__["_signal_"+crc_signal].unsubscribe(subscriber_obj)
                if signal.dbc.attributes[sigType].value == 1:
                    snc_signal = signal.name
                    for subscriber_obj in self.__dict__["_signal_"+snc_signal].subscriptions:
                        self.__dict__["_signal_"+snc_signal].unsubscribe(subscriber_obj)
        data_ids = metadata.dbc.attributes['DataIds'].value
        try:
            data_ids = ast.literal_eval(data_ids)
        except :
            logging.error("can not convert data ids to list")
        self.add_data_ids(data_ids, crc_signal, snc_signal)

    def _update_e2e(self, msg: can.Message) -> None:
        """
        Updates the E2E protection of the message.

        Args:
        - msg (can.Message): The CAN message to be updated.
        """
        seq_counter = self.__dict__["_signal_"+self.e2e.snc_signal_name].get()
        data_id = self.e2e.data_ids[seq_counter]
        crc = compute_profile2_crc(self._can_message.data, data_id )
        self.__dict__["_signal_"+self.e2e.crc_signal_name]._state = crc
        self._can_message.data[0] = crc
    
    def _update_can_message(self, value) -> None:
        """
        Updates the CAN message.
        """
        self._update_message()    
        self._update_e2e(self._can_message)
        if self.running:
            logging.debug("update periodic e2e send message {}".format(self._metadata._name) )
        else:
            logging.debug("periodic e2e send message {} is not started".format(self._metadata._name))

    async def __cycle(self, value):
        """
        Sends the periodic message and returns the SNC.

        Args:
        - value: The value to be passed to the coroutine.

        Returns:
        - snc (int): The updated sequence number counter.
        """
        self.timestamp = datetime.datetime.now().timestamp()
        self.send()
        logging.debug("send periodic e2e send message {} at timestep {}".format(self._metadata._name, self.timestamp))
        await asyncio.sleep(self._metadata.cycle_time / 1000.0)
        snc = self.__dict__["_signal_"+self.e2e.snc_signal_name].get()
        return snc

    def _increase_counter(self, snc:int):
        """
        Increases the SNC and notifies subscribers.

        Args:
        - snc (int): The current sequence number counter.
        """
        self.__dict__["_signal_"+self.e2e.snc_signal_name].notify( (snc + 1) % len(self.e2e.data_ids) )
        self.notify(self._can_message)

    def start_periodic(self):
        """
        Starts the periodic message.
        """
        if self._metadata.cycle_time:
            for signal in self._metadata._signals:
                self.__dict__["_signal_"+signal.name].init_value()
            self._periodic_publisher = (self | op.MapAsync(self.__cycle, mode=op.AsyncMode.CONCURRENT))
            self._periodic_feedback = self._periodic_publisher.subscribe(Sink(self._increase_counter))
            logging.debug("set periodic publisher for message {}".format(self._metadata._name))
            self.running = True
            self._update_can_message(None)
        else:
            print("is not a periodic message")

    def stop_periodic(self):
        """
        Stops the periodic message.
        """
        if self._periodic_feedback is not None:
            #self._periodic_publisher.dispose()
            self._periodic_feedback.dispose()
            self.running = False
        else:
            print("is not a periodic message")


class MessageRx(Message):
    """
    A class representing a received CAN message.

    Inherits from the Message class and adds functionality for decoding and updating signals.

    Attributes:
    - metadata (database.Message): The metadata for the message.
    - timestamp (float): The timestamp for the message.
    - _signal_* (Signal): The signals for the message.
    - _get_* (method): Methods for getting the value of each signal.
    - _set_* (method): Methods for setting the value of each signal (not available for receiver signals).
    - e2e (EndToEnd): The end-to-end configuration for the message (if applicable).

    Methods:
    - __init__(self, metadata:database.Message): Initializes the MessageRx object.
    - _update_data(self, msg: can.Message) -> None: Updates the signals for the message based on the received CAN message.
    """
    def __init__(self, metadata:database.Message):
        super().__init__(metadata)
        self.timestamp = 0
        for signal in metadata._signals:
            setattr(self, "_signal_"+signal.name, Signal(signal) )
            setattr(self, "_get_"+signal.name, types.MethodType(lambda self: self.__dict__["_signal_"+signal.name].get() ,self ) )
            self.__dict__["_signal_"+signal.name].subscribe(Sink( lambda value: logging.debug("Signal {} changed to value = {}".format(signal.name, value) ) ))
            setattr(self, "_set_"+signal.name, types.MethodType(lambda self,value:  logging.error("is a resiver signal and can not be set! value={}".format(value)),self ) )
            sigType = 'GenSigFuncType'
            if sigType in signal.dbc.attributes:
                if signal.dbc.attributes[sigType].value == 2:
                    crc_signal = signal.name
                if signal.dbc.attributes[sigType].value == 1:
                    snc_signal = signal.name
        if "DataIds" in metadata.dbc.attributes:
            data_ids = metadata.dbc.attributes['DataIds'].value
            try:
                data_ids = ast.literal_eval(data_ids)
                self.add_data_ids(data_ids, crc_signal, snc_signal)
            except :
                logging.debug("MSG rx is not with E2E, can not convert data ids to list")

    
    def _update_data(self, msg: can.Message) -> None:
        """
        Updates the signals for the message based on the received CAN message.

        Args:
        - msg (can.Message): The received CAN message.
        """
        try:
            data_decoded = self._metadata.decode(msg.data)
        except ValueError:
            logging.ERROR('Received unknown message with arbitration id {}'.format( msg.arbitration_id ) )
        logging.debug('Update signals: %s', self._metadata._name)
        if self.e2e is not None:
            seq_counter = data_decoded[self.e2e.snc_signal_name]
            data_id = self.e2e.data_ids[seq_counter]
            crc = compute_profile2_crc(msg.data, data_id )
            if crc != msg.data[0]:
                logging.error('CRC check failed for message %s', self._metadata._name)
                return
        self.timestamp = msg.timestamp
        for signal_name, signal_value in data_decoded.items():
            self.__dict__["_signal_"+signal_name].notify(signal_value)


def create_message(metadata:database.Message, sender:bool, _can_bus:can.BusABC=None ) -> Message:
    """Create a message (Value)"""

    if sender:
        #if metadata.cycle_time and metadata.signals[0].dbc.attributes['GenSigFuncType'].value == 2:
        if metadata.cycle_time and 'DataIds' in metadata.dbc.attributes:
            msg_obj= MessageTxCycleE2E(metadata, _can_bus)
        elif metadata.cycle_time:
            msg_obj= MessageTxCycle(metadata, _can_bus)
        else:
            msg_obj= MessageTx(metadata, _can_bus)
    else:
        msg_obj= MessageRx(metadata)
    return msg_obj


def create_message_class(name: str, is_tx: bool, signals: list[str], messages: list[str]) -> type[Message]:
    """
    Creates a dynamic class definition for messages RX and TX with messages and signals as setter and getter property.

    todo: this is not used yet and not tested

    Args:
    - name (str): The name of the class.
    - is_tx (bool): True if the class represents a TX message, False if it represents an RX message.
    - signals (List[str]): The list of signal names for the message.
    - messages (List[str]): The list of message names for the message.

    Returns:
    - Type[Message]: The dynamically generated class definition.
    """
    # Define the base class
    base_class = MessageTx if is_tx else MessageRx

    # Define the class dictionary
    class_dict = {
        # Constructor
        "__init__": lambda self, metadata: base_class.__init__(self, metadata),

        # Signals
        **{f"_{signal}": property(lambda self: self._get_signal(signal), lambda self, value: self._set_signal(signal, value)) for signal in signals},

        # Messages
        **{f"_{message}": property(lambda self: self._get_message(message), lambda self, value: self._set_message(message, value)) for message in messages}
    }

    # Create the class
    message_class = type(name, (base_class,), class_dict)

    return message_class



from setuptools import setup, find_packages

with open('README.md', 'rb') as readme_file:
    readme = readme_file.read().decode('utf-8')

setup(
    name='canbro',
    version='0.0.2-alpha',
    description='This package extend the python-can with the broqer package. This provides the functionality to work in a reactive style with can signals and messages.',
    author='Lukas Riegler',
    author_email='lukasantonriegler@gmail.com',
    url='https://github.com/lki1354/python-canbro',
    packages=find_packages(),
    long_description_content_type='text/markdown',
    long_description=readme,
    install_requires=[
        'python-can>=3.3',
        'cantools>=36.2',
        'broqer>=1.7',
    ],
    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ],
)



import asyncio
from broqer import Value, op, Sink
v = Value()
throttle_publisher = v | op.Throttle(0.1)
_d = throttle_publisher.subscribe(Sink(print))
v.emit(1)

v.emit(2)
asyncio.get_event_loop().run_until_complete(asyncio.sleep(0.05))
v.emit(3)
asyncio.get_event_loop().run_until_complete(asyncio.sleep(0.2))

# It's also possible to reset the throttling duration:
v.emit(4)

v.emit(5)
asyncio.get_event_loop().run_until_complete(asyncio.sleep(0.05))
throttle_publisher.reset()